# NaifJs, a simple state-machine based dialog manager

NaifJs is a simple state-machine based (task-oriented) [dialog manager](https://en.wikipedia.org/wiki/Dialog_manager) SDK for nodejs developers.

- [Stateful multi-turn dialogues](#stateful-multi-turn-dialogues)
- [NaifJs overall architecture](#naifjs-overall-architecture)
- [Dialogue as a state-machine](dialogue-as-a-state-machine)
- [Run-time state tracker](#run-time-state-tracker)
- [Dialog Units](#dialog-units)
- [Coding a dialogue with a javascript domain specific language](#coding-a-dialogue-with-a-javascript-domain-specific-language)
- [A dialog application](#a-dialog-application)
- [Sessions](#sessions)
- [Installation](#installation)
- [API and command line tools](#api-and-command-line-tools)
- [Development Workflow](#development-workflow)
- [Discussion](#discussion)

## Stateful multi-turn dialogues

A dialogue is a conversational exchange between two or more people, speaking some natural language. 
In a human-computer interaction context, dialogue takes place between two systems: 
the user (A) and a computer program B, usually called *chatbot* or *conversational agent* 
(I prefer to say *conversational application*).

Consider this simple turn-taking:

> A > what time is it?<br>
> B < it's half past seven.<br>

It's a dialogue? 
In a broad sense yes, but it's a special case of conversation because it finish after just a single back and forth exchange.
I call *0-state dialogue* this *single-turn* scenario, because the conversation is apparently *stateless*. 

So I prefer to define as dialogue a *multi-turn* *stateful* turn-taking between two systems.

> A > Is it expected to rain today?<br>
> B < There’s no rain today, there’s sunshine and the temperature is 18 degrees centigrade.<br>
>
> A > Well, how about tomorrow?<br>
> B < It will rain tomorrow.<br>

In the exchange of sequences here above, the topic of the conversation concerns the weather forecast. 
The initial question of system A (the user) brings system B (the chatbot) to an *internal state*, 
which maintains the topic of discussion, so that the subsequent user request is understood in the specific context "weather forecast". 
This is an example of elementary multi-turn conversation (I call *state-1*).

It is therefore possible to model more complex dialog as  a finite state machine,
where two interacting systems (for example A is a human being and B is a chatbot), 
evolve into successive conversational states.


## NaifJs architecture 

The general architecture of a dialog system based on NaifJs is shown in figure 1. 

| ![dialogue system figure 1](doc/img/dialogSystem.png) |
|:--:|
| figure 1: Dialogue system architecture, using NaifJs |

We consider a multiuser scenario where many users can interact the dialogue system. 
The main software components are:

- The channel 

  It's where users interact. It could be by example an instant messaging platform, a group chat platform, an email client, etc. 

- The messages data adapter

  The outside interface component of the main program is the channel API that allow to send and receive messages to the channel. 
  ASR (speech transcript / text-to-speech could be part of this component if messages include voice.
  Different types of multimodal messages could be produce inbound `request`s to the dialogue system. 
  Backwards, dialog system `response`s are translated in specific outbound message types.

- The dialog selection/activation 
 
  NaifJs is built on the idea that user interacts with the dialog system activating, one at a time, a dialog flows, 
  I call *dialogue units* (or *dialogue threads*) each modeled by a dedicated state machine. 
  One can see a dialog unit as a micro-skill or workflow to accomplish some task.

  - *Pull events*

    The core logic of the application program must so contain a dialogue selector that activates a dialogue unit (*pull mode*). 
    A common case is that a dialog unit is triggered if a specific *invocation sentence* is matched. 

  - *Push events*

    The invocation sentence is not the only way to `start` a dialogue unit. 
    Another event that can start the dialogue is an event generated by some backend logic. 
    It's common to push user with notification-based conversations, 
    maybe following agent-initiative-based logic.
 
- The dialogue manager
 
  The NaifJs dialog manager "up-API" start a dialogue and pass requests to underlying dialog units.
 
- The application dialogues

   An application (or project) is made by a set of (standalone or interconnected) dialogue units.

- The dialog units

  A dialog unit is just a finite state machine (FSM), 
  a single *unit* of conversational logic to accomplish some specific application workflow.



Naifjs is based on a specific conceptual model and a domain specific language (DSL) I conceived, 
regarding how to implement a dialog unit, following a state machine paradigm. 
Following paragraphs introduce main concepts and naming conventions.

 
## Dialogue as a state-machine

NaifJs is a dialogue manager that implements dialog management modeling a conversation as a finite state machine 
([FSM](https://en.wikipedia.org/wiki/Finite-state_machine)).

A conversation in NaifJs can be thought of as traversing a the state machine, 
to be considered as a graph of *states* (elaboration *nodes*).

### State representation

Each *node* (or *state*) of the state machine has an associated unique identifier (*stateid*), 
a string label identifying the state machine (the dialogue unit) and the node inside the state):

```
'myDialogueUnit.myState'
``` 

Where:

- `myDialogueUnit` is the name of the state machine, the dialogue unit file `myDialogueUnit.js` 
- `myState` is the name of the javascript function handler `myState()` 

### Input and output states

In each state machine there are two different kind of nodes:

- *input state*
 
  It's a "wait" node of "understanding" and local processing of input requests (user *utterances*). 
  The inbound user sentence is matched with some kind of [pattern matching](doc/patternmatching.md) 
  (e.g. Regular Expressions) and each matched "intent" triggers an *action* (any backend logic). 
  Lastly, a response message (*prompt*) is feed back to user. See nodes *S1* and *S2* in figure 2.
  The input state is triggered with `next('myDialogueUnit.myInputState')` DSL.

- *output state* 
 
  It's an execution node that makes some backend elaboration and produce a message/prompt (response) to the user. 
  We can consider the output state as the *action* part of previous described input state. 
  The output state is triggered with `exec('myDialogueUnit.myOutputputState')` DSL.


### Rule-based state transitions

State transitions are deterministic and "rule-based", 
being programmed by developer, in the action part of [pattern matching](doc/patternmatching.md) rules. 

A dialogue is initially activated by a *caller* program (with `start` statement) 
that activate the *callee* state machine. 
Usually (but not perforce) the first state is an (output) state.

The node performs some kind of logic (L, in figure 2), usually (but non perforce) 
feedback a response message (prompt) to the user 
and can store contextual (shared) memory variables (M). 

The response to the user is any kind of message 
(can be a text, a voice message, an image or audio/video content).

| ![state machine figure 2](doc/img/state-machine-1.png) |
|:--:|
| figure 2: at the left a *1-state* *dialogue unit*, containing an *output state*, an *input state*, a local memory. At the right a *2-state* dialogue unit, containing 2 input states, an input state, a local memory. |


## Run-time state tracker

The engine is just a *conversation state* scheduler/tracker. 

As soon a dialogue state machine is activated from a main program (the *caller* or *main state*), 
the NaifJs engine track the conversation state (*stateid*) and schedule next state.

Incoming user request event (the user utterance as text), run a state handler function.

```javascript
// dialogue manager set a specific state-machine state
naif.start('dialogueUnit.dialogueState')

// when user talk to the agent:
naif.request(user, utterance)
```

To progress the dialogue flow, the initial (output) state usually send a prompt to user 
and activates a subsequent input state (using `next` DSL statement), 
in which the dialogue manager analyzes the user's response (a sentence in natural language)
and performs a contextual interpretation based on a [pattern matching](doc/patternmatching.md) (PM). 

The dialogue manager is agnostic with respect
to the NLU mechanism and for simplicity and efficiency
the parsing has been implemented with the use of regular expressions (regexp). 

In the case of a positive pattern match, 
the input processing node performs some processing and a response is given to the user.
The *prompt* could be textual, a synthetic voice speech or other multimedia audio/video content. 

The current state can evolve into a successive input state (using `next` API statement), 
a successive output state (using `exec` DSL statement) or might not change 
(*loopback*, see the dotted lines in figure 3),
based on a dialogue flow programmed by the designer. 

The dialogue flow finally ends, when developer terminate the conversation, 
terminating the state machine (with `end` DSL statement).

| ![state machine figure 3](doc/img/state-machine-1-new.png) |
|:--:|
| figure 3: at the left, the *1-state* dialogue unit, where the input state is detailed. |


## Dialog Units

To complete a specific task (a workflow), 
a multi-turn conversation is implemented as a network of micro-dialogues, 
called *dialogue units*, each containing a graph of nodes that completes a *contextual* 
(related to in a specific domain and task) conversation workflow. 

A dialog unit is a finite state machine (FSM), 
a single *unit* of conversational logic allowing accomplish any specific application workflow.

From developer point of view, each dialogue unit is a standalone file, 
exporting a module that implement a single state machine, or micro-dialogue, 
that has an entry point (the initial state) and one or more ends.

These dialogue units modules could be concatenated each other, resulting a complex dialogue. 
Each dialogue unit can receive input parameters, following a sort of functional programming model. 


| ![state machine figure 4](doc/img/state-machine-2-new.png) |
|:--:|
| figure 4: a more complex state machine, made by 2 interconnected dialogue units. |

**Dialogue Data**

Each dialogue unit has a *short-term memory*: temporary variables with local visibility to the unit itself.

In NaifJs each state doesn't contain any local storage, 
because each node is implemented just as a *stateless* javascript function handler. 

Currently variables can be shared among nodes part of the same dialogue unit (dialogue unit scoped). 
Variables are managed with specific NaifJs API primitives and stored in the [*sessions* database](doc/session.md).


## Coding a dialogue with a javascript domain specific language

To write a *dialogue unit* in NaifJs, the developer just write vanilla javascript code, 
added by NaifJs [DSL statements](doc/API.md#down-dialogue-api) 
that define how input and output states (nodes) of the state machine are interconnected. 

Each state machine correspond to a dialogue unit, an unique javascript module file 
containing a list of functions (handlers) for each *input state*, 
and *output state* of the designed state machine.

### Dialog unit example

Suppose we want to implement a simple dialogue unit 
(e.g. the file `getUserName.js`) that ask the user his name, and terminates. 

Here below the basic code. The example dialogue is a 1-state, 
because it contains just a single input state handler (`getname`). 

As soon the dialogue is activated (with `naif.start('getUserName.askName')`), 
the output state `askName` handler is triggered. 
The output state simply prompt the user and goto the input state with `next(getName)`.

```javascript
// dialog unit: getUserName.js

const getUserName = { askName, getName }
  
// output state handler, dialogue entry point 
function askName() {
  say('Hello! What\'s your name?') // say() is an alias for response()
  next('getUserName.askName')
}

// input state handler 
function getName(sentence) {

  // regexp rule to get a first name (e.g. Anna Lisa, Giorgio La Malfa)
  const firstNameReg = '(?<firstName>[A-Z][a-z]{1,}(\\s[A-Z][a-z]{1,})*)'
  const introNameReg = '^(my name is |name me |i\'m )?'
  const FIRSTNAME_REGEX = new RegExp ( introNameReg + firstNameReg, 'i' ) 
  const HELLO_REGEXP = /hi|hello|good morning|good afternoon/i

  switch (take(sentence)) {

    case match( HELLO_REGEXP ): // pattern 1
      say('Hi there! What\'s your name?')
      break
   
    case match( FIRSTNAME_REGEX ): // pattern 2
      say(`Nice to meet you, ${slots().firstName}?`)
      end()
      break
     
    default: // fallback
      say( 'I do not understand.' )
      say( 'What is your name?' )
      break
  }
}

module.exports = getUserName 
```


## Dialog application

In NaifJs, a *dialogue application*, or *project* is a set of dialogue units files, 
currently contained in a single specified directory. 
At run time all dialogue files of the directory are loaded into the NaifJs engine, 
with DSL statement `up`:

```javascript
// myApplicationMain.js
const naif = require('naifjs')
const DIALOGS_DIRECTORY = './myApplicationDialogues'
naif.up(DIALOGS_DIRECTORY, responseCallback)
naif.start('firstUnit.start')
```

The application caller invoke the node `start` of dialogue unit `firstUnit.js`. 
Afterward the user will traverses the state machine graph, 
maybe passing from dialogue unit `firstUnit.js` to dialogue unit `secondUnit.js`.

```
path/to/myProject
├── myApplicationMain.js 
└── myApplicationDialogues 
  ├── firstUnit.js
  └─── secondUnit.js
```

In general, an application could be considered as a container of one or multiple dialogue units (micro-skills),
each implementing a single self-consistent workflow.

| ![complete application figure 5](doc/img/main-state.png) |
|:--:|
| figure 5: a complete application,  made by 4 dialogue units. |


### Application example (Telegram Bot) 

The example here below implement a simple Telegram chatbot that invoke the dialogue 'greetings', 
when the user say the *activation sentence*: `greetings`. 

The invocation sentence is captured by a regexp that optionally catch also a parameter (slot) 
passed to the dialogue entry function handler. 

```javascript
const TelegramBot = require('node-telegram-bot-api')
const naif = require('naifjs')

const bot = new TelegramBot('YOUR_TELEGRAM_BOT_TOKEN', {polling: true})

// directory containing all dialog-units
const PROJECT_DIRECTORY = './greetings'

const GREETINGS_INVOCATION_REGEXP = /^greetings\s*(.+)*$/i 
const GREETINGS_DIALOG_ENTRY = 'getUserName.askName'

naif.up(PROJECT_DIRECTORY, bot.sendMessage)

// process any incoming message
bot.on('message', msg => {

  let match

  const userId = msg.chat.id
  const userSentence = msg.text

// There is a dialogue already active for the user?
if ( naif.isSessionActive(userId) )
  naif.request(userId, userSentence)

// dialog-unit start when the invocation sentence is matched
if ( (match = userSentence.match(GREETINGS_INVOCATION_REGEXP)) )
  naif.start(userId, GREETINGS_DIALOG_ENTRY, match[1])

// fallback answer
bot.sendMessage(userId, 'Sorry, I do not understand.') 
})

process.on('SIGTERM', naif.down())
```

### Requests and responses

NaifJs interact with user receiving and sending messages. 

The platform is agnostic regarding media contents in `request` and `response` messages. 
A messaging channel adapter module (external to NaifJs) 
is in charge to transfer inbound messages supplying and request data object to the `request` API. 

The adapter is also in charge to translate the outbound `response` data object in a message type 
appropriate to the channel surface. 
E.g. the response could be a text or a speech or an image to be displayed on the channel 
(by example a mobile instant messaging chat app).
Read more [here](doc/requestresponse.md).

NaifJs is able to manage the user silence (no answer) after a prompt. 
This is done associating a timeout handler to a `response` function. 
For details see [Response Timeout Handlers](doc/requestresponse.md#response-timeout-handlers).


## Sessions 

NaifJs is a multi-user dialogue manager. 

Even if running as an embedded library, the engine can be used as a "server" that manage dialogues by many *user ids*, in parallel. 
More details [here](doc/sessions.md).


## Installation

The package contains command line interface program `naif`, so you must install the npm package as global:

- use npm package manager repo

```
$ npm install -g naifjs
```

- or download this github repo:

```
$ git clone https://github.com/solyarisoftware/naifjs
$ cd naifjs && npm link
``` 

## API and command line tools

### API

Currently NaifJs act as an API library of functions to be called by a main nodejs program.
See [API](doc/API.md) documentations and examples. 

### Command line interface utilities

The CLI program [`naif`](doc/CLI.md) act as a command line tester/debugger, 
a code generator and it supply some application info.

- [`naif show`](doc/CLI.md#naif-show) - Application dialogues directory info
- [`naif test`](doc/CLI.md#naif-test) - How to test dialogues on the command line
- [`naif telegram`](doc/CLI.md#naif-telegram) - How to test dialogues with a Telegram bot
- [`naif new`](doc/CLI.md#naif-new) - A basic dialogue code skeleton generator

### Project examples

You can explore a bunch of NaifJs dialogue unit examples [here](/examples).


## Development Workflow

A discussion about the design/developement [workflow](doc/workflow.md).


## Discussion 

- [Pros and Cons](doc/discussion.md#pros-and-cons)
- [BackStory](doc/discussion.md#backstory)
- [To do](doc/todo.md)


## How to contribute / To do

Please feel free to contact me for any proposal/discussion. Here [how to contribute](CONTRIBUTING.md).<br>
There are many areas of improvement, see the [to do list](doc/todo.md).


## ⚠️  Disclaimer

Currently the project is just a proof of concept. Not ready for production.
The software made is just a prototype in alpha stage. 


## License 

[MIT](LICENSE.md)

---

[top](#) | [home](../README.md)
