const os = require('os')
const readline = require('readline')

const { up, down, start, request, banner } = require('../../')
const { dialogsAndStates } = require('../../lib/unitsLoader')
const { version, author } = require('../../package.json')

/*
 * 39 Default foreground
 * 30 Black
 * 31 Red
 * 32 Green
 * 33 Yellow
 * 34 Blue
 * 35 Magenta
 * 36 Cyan
 * 37 Light gray
 * 90 Dark gray
 * 91 Light Red
 * 92 Light Green
 * 93 Light Yellow
 * 94 Light Blue
 * 95 Light Magenta
 * 96 Light Cyan
 * 97 White
 */
const COLOR_ADMIN = 36
const COLOR_ERROR = 91
const COLOR_BOT_WRITE = 92 // 36
const COLOR_BOT_TTS = 93 // 36


/**
 * naif test logo
 * @see  http://patorjk.com/software/taag/#p=testall&f=Blocks&t=naif test 
 */
function logo() { 
  return( `naif test. command line dialog tester
version: ${version}. Author: ${author.email}
 `)
}
   

/**
 * naiftest usage info 
 */
function usage(errorString) {
  
  console.log( banner() )
  console.log('Command line dialog tester') 
  console.log()
  console.log('Usage:\n    naif test\n')
  console.log('    --dir=<directory path>')
  console.log('      Dialog units directory path. Mandatory')
  console.log()
  console.log('    --start=<unit.state>')
  console.log('      Initial dialog stateid in format "unit.state". Mandatory')
  console.log()
  console.log('    --id=<id>')
  console.log('      User id. Optional argument')
  console.log()
  console.log('    --logfile=<path/filename>')
  console.log('      Log filename with directory path. Optional')
  console.log()
  console.log('    --sessionsfile=<path/filename>')
  console.log('      Sessions filename with directory path. Optional')
  console.log()
  console.log('Examples:')
  console.log()
  console.log('   naif test --start=firstUnit.start --dir=examples/myDialogs')
  console.log()
  console.log('   naif test --dir=examples/app \\')
  console.log('             --start=anotherUnit.showContents \\')
  console.log('             --id=123456 \\')
  console.log('             --logfile=examples/app/dialogs.log \\')
  console.log('             --sessionsfile=examples/app/sessions.json\n')

  if (errorString) 
    console.log(`\n\n  ${errorString}\n`, COLOR_ERROR)

  process.exit()
}


/**
 * validate command line arguments 
 */
function checkArgs(args) {

  // no arguments
  if ( Object.entries(args).length === 0 && args.constructor === Object ) {
    usage()
    process.exit()
  }

  // dialogs directory
  if ( ! args.dir ) {
    usage('ERROR: dir not specified.')
    process.exit()
  }  


  // default start state
  if ( ! args.start) {
    usage('ERROR: starting stateid not specified.')
    process.exit()
  }  

}


/**
 * console log (coloured) 
 *
 * @param {String} sentence
 * @param {Integer} color code
 * 
 * @see https://misc.flogisoft.com/bash/tip_colors_and_formatting
 *
 */ 
function writeTerminal(text, color=COLOR_ADMIN) {
  text = (text) ? text : '' 
  console.log(`\x1b[${color}m${text}\x1b[0m`)
}

/**
 * print to stdout the message to the user 
 *
 * @param {String} userid 
 * @param {String/ResponseDataObject} data
 *
 */ 
function response(userid, data ) { 

  // userid is not used / not useful here
  writeTerminal( '\n' + data.text, (data.speech) ? COLOR_BOT_TTS : COLOR_BOT_WRITE)
  //log.out(userid, `[${engine.getStateId()}] ${sentence}`)
}


/**
 * end the conversation 
 */ 
function end() {
  const text = '\n' + 'dialog end.'
  writeTerminal(text)

  // shutdown the engine
  exit() 
}


/**
 * shutdown naif engine and exit from current process.
 */ 
async function exit() {

  statistics()

  // shutdown naif engine
  await down()

  // exit from current process
  const text = 'naif engine shutted down.'
  console.log(text)
  
  process.exit(0)
}  


/**
 * get os username 
 */
function loginUserid() {
  const username = os.userInfo().username
  const hostname = os.hostname()
  return(`${username}@${hostname}`)
} 


/**
 * check if a string is uppercase
 *
 * @param {String} str string to be check
 * @return {Boolean} true if the string is uppercase 
 * 
 * @example isUpperCase("hello"); // false
 * @example isUpperCase("Hello"); // false
 * @example isUpperCase("HELLO"); // true
 *
 */
function isUpperCase(str) { // eslint-disable-line no-unused-vars
  return str === str.toUpperCase()
}


/**
 * check if sentence is spoken or written
 * to simulate a spoken sentence, check if it is uppercase
 * @param {String} str string to be check
 * @return {Boolean} true if the string is spokene 
 */
function isSpoken(str) {
  return str === str.toUpperCase()
}


function help() {
  writeTerminal()
  writeTerminal('  Help:')
  writeTerminal()
  writeTerminal('  An utterance made by uppercased words emulate voice (speech) utterance:')
  writeTerminal()
  writeTerminal('    hello Giorgio -> is a chat (written) utterance')
  writeTerminal('    HELLO GIORGIO -> is considered a spoken utterance. Tip: use caps-lock key.')
  writeTerminal()
  writeTerminal('  Use Up-Arrow and Down-Arrow keys to scroll and retrieve previous utterances.')
  writeTerminal()
  writeTerminal('  CTRL-C to exit.')
  writeTerminal()

}


function statistics() {
  const wpu = (words/cpsIn).toFixed(2)

  writeTerminal()
  writeTerminal('Session Statistics:')
  writeTerminal()
  writeTerminal('Conversation Entry Point (stateid)   : ' + startStateId)
  writeTerminal('CPS (Conversation-turns Per Session) : ' + cpsIn)
  writeTerminal('CPS Written Utterances               : ' + (cpsIn - cpsInSpoken) )
  writeTerminal('CPS Spoken Utterances                : ' + cpsInSpoken)
  writeTerminal('WPU (Words Per Utterance)            : ' + ( isNaN(wpu)? 0 : wpu ) )
  writeTerminal('Elapsed Time (secs)                  : ' + (elapsedTime(startTime)/1000).toFixed(2) )
  writeTerminal()
}



//
// elapsed time 
//
function startTimer() {
  return new Date()
}


function elapsedTime(startTime) {
  const endTime = new Date()
  let timeDiff = endTime - startTime //in ms
  return timeDiff
}


/**
 * @see https://stackoverflow.com/a/18679657/1786393
 */  
function wordCount(str) {
  return str.split(' ')
    .filter(function(n) { return n != '' })
    .length
}


let firstTimeHelp = true
let cpsIn = 0
let cpsInSpoken = 0
let words = 0
let startStateId
let startTime


/*
 * naif test main
 */ 
function naiftest(args) {

  checkArgs(args)

  const id = (args.id) ? args.id : loginUserid()
  const UnitsDirectory = args.dir
  startStateId = args.start
  const Logfilename = args.logfile
  const sessionsfilePrompt = args.sessionsfile? args.sessionsfile : 'not specified -> ./sessions.json'

  console.log( logo() )

  // NaifJs dialog engine setup
  // load units from specified folder
  up( UnitsDirectory, response, end, args.sessionsfile, Logfilename, false )

  console.log('\nDialog Units Directory\n' + UnitsDirectory)
  console.log('\nDialog States')
  console.log(dialogsAndStates().join('\n'))
  console.log('\nStart Dialog\n' + startStateId)
  console.log('\nSessions File\n' + sessionsfilePrompt)
  console.log('\nUser\n' + id)

  if (Logfilename)
    console.log('\nLog File\n' + Logfilename)

  //
  // Command line events handling
  //
  const cli = readline.createInterface({
    input: process.stdin,
    output: process.stdout,
    prompt: '' //'>> '
  })

  // https://misc.flogisoft.com/bash/tip_colors_and_formatting
  cli.on('close', () => {
    const text = 'Shutdown (session closed).'
    console.log( '\n' + text)
    exit() 
  })

  //
  // Ctrl-C
  //
  cli.on('SIGINT', () => { 
    const text = 'Shutdown (Control-C).'
    console.log( '\n' + text)
    exit() 
  })


  start(id, startStateId)

  startTime = startTimer()

  // process stdin line
  cli.on( 'line', sentence => { 
    
    sentence = sentence.trim()

    if (sentence) {
      
      cpsIn ++
      words += wordCount(sentence)

      const spokenFlag = isSpoken(sentence) 

      if (spokenFlag) {
        cpsInSpoken ++
        request(id, {text: sentence, speech: true})
      }  
      else
        request(id, {text: sentence, speech: false})
    }  
    else {
      // no sentence (user press enter)
      writeTerminal('Write an utterance to continue.')
      if (firstTimeHelp) {
        help()
        firstTimeHelp = false  
      }  
    }  

  })
}

module.exports = { naiftest }
