# Requests and responses message types

NaifJs is a backend dialog manager engine, multi-user, channel agnostic.

Each inbound and outbound messages are to/from a specific *sessionid* (the user id).
NaifJs is independent from the frontend messaging channel
(that could be an instant messaging app like Telegram,
Slack, etc., or a command line interface, anything).


- Request (inbound) messages
 
  The main program, when calling NaifJs API through a specific messaging adapter API, 
  is in charge to get messages from the channel.

- Response (outbound) messages 

  The main program, when calling NaifJs API through a specific messaging adapter API, 
  is in charge to send messages to the channel, using different media content types.


## Request message data 

Using NaifJs, the main node program is in charge to convert each incoming message from the channel, 
calling the dialog manager with `request` API.

NaifJs accepts requests data payload as:

- flat text 

- a javascript object data structure 

  containing at least a `text` attribute,(denoting a user written utterance), 
  a `speech` attribute, denoting that user spoken the utterance, 
  and more arbitrary optional parameters.

Examples:

- Written input, where data payload is a string):

  ```javascript
  naif.request(sessionid, 'my name is Giorgio Robino' )
  ```

- Written input, where data payload is an object:

  ```javascript
  naif.request(sessionid, {text: 'my name is Giorgio Robino'} )
  ```

- Speech (transcript from a speech-to-text engine), where data payload is an object:

  ```javascript
  naif.request(sessionid, {text: 'my name is Giorgio Robino', speech: true} )
  ```


## Response message data

The same main program has to supply a `response` callback to send the message to the channel 
(see the Telegram bot pseudo-code example here above).
The output message (reply to user) could be :

- a flat text (attribute of DSL statement `response`)
- a specific javascript object data structure. 

Here below some examples with different kind of `response` data payloads:

- text only output:

  ```javascript
  naif.response(sessionid, { text: 'my name is Giorgio Robino' } )
  ```

- speech with text caption:

  ```javascript
  naif.response(sessionid, { text: 'my name is Giorgio Robino', speech: true, caption: 'my name is Giorgio Robino' } )
  ```

- image with text caption and voice over:

  ```javascript
  naif.response(sessionid, { 
    image: 'path/to/image.jpg', 
    text: 'my name is Giorgio Robino', 
    speech: true, 
    caption: 'my name is Giorgio Robino' 
  } )
  ```

- [earcon](https://en.wikipedia.org/wiki/Earcon):

  ```javascript
  naif.response(sessionid, { earcon: 'path/to/earcon.ogg', text: false, } )
  ```

There are two types of responses:

- *statements*
 
  The agent response doesn't require a turn-taking; the user may not interact anymore.
  In the example here below, the `response` API is aliased with `say`:

  ```javascript
  say('Hello! Today is sunny, with a temperature of 15 degrees')
  ```

- *prompts*
 
  The response contains a question for the user who is asked to reply something.
  In other words, the agent ask a question to the user.


### Prompt timeout handler

A dialog agent prompt to user with a `response` API statement. 
By default there is no any timeout management: 

```javascript
say('Hello! What\'s your name?')
```

In the case above the dialog is stuck, 
because the agent is waiting indefinitely a reply from the user.
That's a bad practice and it could be avoided.

NaifJs is able to manage the user silence (no answer) after a prompt.
This is done passing an optional object to the `response` function:

```javascript
response('Hello! What\'s your name?', { callback: askNameTimeout, milliseconds: DELAY_15S })
```

The timeout callback is triggered after a certain number of specified milliseconds 
and usually it just reprompt the user. 
In the example here below, the handler reprompt twice, afterward give-up. 

```javascript
// output state handler, dialog entry point (start) 
function askName() {
    
  say('Hello! What\'s your name?', askNameTimeout)

  next(getNameState)
}

// input state handler 
function getName(sentence) {

  // create a regexp rule to get a first name (e.g. Anna Lisa, Giorgio La Malfa)
  const firstNameReg = '(?<firstName>[A-Z][a-z]{1,}(\\s[A-Z][a-z]{1,})*)'
  const introNameReg = '^(my name is |name me |I\'m )?'
  const FIRSTNAME_REGEX = new RegExp ( introNameReg + firstNameReg ) 

  // the function askNameTimeout is called 15 secs after the say statement 
  const timeout = { callback: askNameTimeout, milliseconds: (15 * 1000) }

  switch ( take(sentence) ) {
 
    // pattern 1
    case match( /hi|hello|good morning|good afternoon/i ):
      
      ask( 'Hi there! What\'s your name?', timeout )
      break

    // pattern 2
    case match( FIRSTNAME_REGEX ):

      say( `Nice to meet you, ${slots().firstName}!` )
      end()
      break

    // fallback pattern
    default:

      say( 'I do not understand.' )
      ask( 'What is your name?', timeout )
      break
  }
}


/**
 * ask name timeout handler
 *
 * @param {string} user id
 * @param {Object} params
 */ 
function askNameTimeout(userid, params) {

  setChat(userid)

  switch( getCounterTimeout() ) {

    case 1:
      // first timeout goes off
      ask('What\'s your name?')

      // reschedule a timeout in 30 seconds
      startTimeout(askNameTimeout, 30000)
      break

    case 2:
      // second timeout
      ask('Are you there? Please tell me what is your name')

      // reschedule a timeout in 60 seconds
      startTimeout(askNameTimeout, 60000)
      break

    case 3:
      // third timeout: close the conversation
      stopTimeout()
      end()
   }   
 }   
```

The running example is the [here](../examples/getUserNameTimeout/getUserName.js).

### Prompt policy 

🗒 Please note that NaifJs allows developer to decide how to do when timeout is up.
In the example above the user is re-prompted twice, 
he's asked to reply, afterward  the dialogue end.
This is an arbitrary policy. 

Another option could be to let the dialog move to another state, 
maybe reformulating the question or moving to another topic.

---

[top](#) | [home](../README.md)
