const log = require('./log')
const { banner } = require('../bin/lib/info')
const { getSessionId, setSessionId } = require('./sessionid')
const { validateRequestData } = require('./messaging/validateRequestData')
const { stopTimeout } = require('./timeouts')
const { setExternalResponse } = require('./externalResponse')

const { startTimeout } = require('./timeouts')
const { validateTimeout } = require('./messaging/validateTimeout')
const { validateResponseData } = require('./messaging/validateResponseData')
const { getExternalResponse } = require('./externalResponse')

// dialog handlers loader
const { Dialogs, loadUnits, printStates } = require('./unitsLoader')

// sessions
const { 
  isSessionActive,
  getSession, 
  setSession, 
  delSession, 
  getSessionVar, 
  setSessionVar, 
  delSessionVar, 
  delSessionVars,
  initSessions, 
  dumpSessions 
} = require('./sessions') // eslint-disable-line no-unused-vars


//const TAGREGEXP = /[.:/]/
const TAGREGEXP = '.'


/*
 * engine log filename.
 * @private
 */
let logFilename


/*
 * spoken/written flag
 * true if sentence is spoken.
 * @private
let asrFlag
 */

/*
 * dialog status as pointer to actual handler function
 * @private
 */
let unit
let state

/**
 * local pointer to external function to end
 * @private
 */
let externalEnd

/**
 * default start stateid 
 * @private
let defaultStartStateid // eslint-disable-line no-unused-vars
 */

/**
 * get current dialog stateid 
 * @return {String}
 */
function stateid() {
  return `${unit}.${state}`
}  


/**
 * set user variable
 *
 * @param {String} name of variable
 * @param {Object} value 
 * @param {String} unit name
 * @param {String} sessionid 
 *
 * @returns {Object} value
 *
 */
function setvar(name, value, unitid=unit, sessionid=getSessionId()) {
  // DEBUG
  //console.log(sessionid, unitid, name, value)
  return setSessionVar(sessionid, unitid, name, value)
}


/**
 * get user session variable
 *
 * @param {String} name name of variable
 * @param {String} unit name of unit
 * @param {String} sessionid 
 * @returns {Object} 
 *
 */
function getvar(name, unitid=unit, sessionid=getSessionId()) {
  // DEBUG
  //console.log(sessionid, unitid, name)
  return getSessionVar(sessionid, unitid, name)
}


/**
 * del user session variable
 *
 * @param {String} name name of variable
 * @param {String} unit name of unit
 * @param {String} sessionid 
 *
 */
function delvar(name, unitid=unit, sessionid=getSessionId()) {
  return delSessionVar(sessionid, unitid, name)
}


/**
 * del all user session variables
 *
 * @param {String} unit name of unit
 * @param {String} sessionid 
 *
 */
function delvars(unitid=unit, sessionid=getSessionId()) {
  return delSessionVars(sessionid, unitid)
}


/**
 * log object as JSON in logfile
 *
 * @param {Any JS Object} data
 * @param {String} text. optional
 * @param {String} sessionid user caht id default is current id
 */ 
function logData( data, text, sessionid=getSessionId() ) {

  if (logFilename) {
    const formattedStateid = `[${stateid()}]`

    // if text is present, prefix the stateid in brackets format
    const outText = (text) ? `${formattedStateid} ${text}` : formattedStateid 

    // if data is present, add the stateid as attribute
    if ( typeof data === 'object') 
      data.stateid = stateid()

    log.data(sessionid, outText, data)
  }  
  else
    console.error(`logData(${data}) called, but logFilename wasn't specified.`)
}


/**
 * End of a conversation
 * Call an external handler
 *
 */
function end(sessionid=getSessionId()) {
  
  // set session not active
  delSession(sessionid)

  externalEnd(sessionid) 
}


/**
 * setup engine
 *
 * @public
 * @param {String} unitsFolder
 * @param {String} response / replyFunction
 * @param {String} stateid / default start state
 * @param {String} aLogFilename  
 * @param {Boolean} [true] verbosity 
 *
 */
function up(unitsFolder, responseCallback, endCallback, sessionsFilename, aLogFilename, verbose=true) {

  if (verbose) 
    console.info(banner())

  // load units
  build(unitsFolder, verbose)

  // assign external callback to reply 
  setExternalResponse(responseCallback)

  // assign external callback to end 
  externalEnd = endCallback

  // assign default start stateid
  //defaultStartStateid = startStateid

  // Load sessions in memory
  // if a valid filename is not passed, 
  // a sessions.json is created in the application directory
  initSessions(sessionsFilename ? sessionsFilename : unitsFolder+'/sessions.json')
  
  if (aLogFilename) {

    logFilename = aLogFilename

    // open dialogs log file
    log.open(logFilename)
  }
}  


/**
 * shutdown engine
 *
 * @public
 *
 */
function down() {

  // dump sessions to disk
  dumpSessions()

  // close dialogs log file
  //log.close()

} 


/**
 * load units
 * get the dialogs lookup map
 *
 * @private
 * @param {String} unitsFolder
 * @param {Boolean} verbose
 *
 */
function build(unitsFolder, verbose) {

  loadUnits(unitsFolder, verbose)
  
  if (verbose)
    printStates()

}


/**
 * start state by stateid
 *
 * @param {string} stateid unit.state string stateid 
 * @param {string} sessionid 
 * @param {AnyObject} argument (any object). It will passed to initial output state (by stateid) 
 */ 
function start(sessionid, stateid, argument) { 

  setSessionId(sessionid)

  // load the sessionid session assigning the initial state stateid
  exec(stateid, sessionid, argument)

}


/**
 * schedule and run an OUTPUT state 
 *
 * @param {string} stateid 
 * @param {string} sessionid user chat id. optional
 * @param {AnyObject} argument (any object). It will passed to initial output state (by stateid) 
 *
 * action
 *   set next dialog handler
 *   pass start argument as output state function argument 
 *   call the dialog 
 *
 */ 
function exec(stateid, sessionid=getSessionId(), argument) {

  next(stateid, sessionid)

  //
  // Call OUTPUT state function handler
  // pass the user id and the arg to the called function 
  Dialogs[unit].unit[state](sessionid, argument)

}


/**
 * schedule next INPUT state
 *
 * @param {string} stateid 
 * @param {string} sessionid user chat id. optional
 *
 * action
 *   prepare next dialog handler
 *   verify if unit.state exist looking at dialogs map
 *   store the status pointer
 *
 */ 
function next(stateid, sessionid=getSessionId()) {
  
  [unit, state] = stateid.split(TAGREGEXP)
  
  //
  // run-time state stateid validation
  //
  if ( !unit || !state ) 
    throw new Error('Wrong format. Please specify both unit and state with format: UnitName.StateName')
  
  if ( ! unitExist(unit) ) {
    const errorMsg = `dialog '${unit}' not found.`
    throw new Error(errorMsg)
  }  

  if ( ! stateExist(unit, state) ) {
    const errorMsg = `state '${state}' of dialog '${unit}' not found.`
    throw new Error(errorMsg)
  }  

  // Update session
  setSession(sessionid, unit, state)
}


/**
 * verify if unit exist in dialogs map
 *
 * @return {Boolean}
 *
 */ 
function unitExist(unit) {
  return Dialogs[unit] ? true : false
}


/**
 * verify if state exist in DialogName, looking up dialogs map
 *
 * @return {Boolean}
 *  
 */
function stateExist(unit, state) {
  return Dialogs[unit].states.includes(state)
}


/**
 * request
 * the user message trigger the the engine,  
 *
 * @param {String} sessionid
 * @param {object} data - request data payload object
 *
 */
function request( sessionid, data ) { 

  const validatedData = validateRequestData(data)

  setSessionId(sessionid)

  // clear any pending timeout action
  stopTimeout(sessionid)

  // trim the input sentence 
  const sentence = validatedData.text.trim()

  if (logFilename)
    log.inp(sessionid, `[${stateid()}]${validatedData.speech? ' [speech] ': ' '}${sentence}`)

  // Retrieve session data
  const session = getSession(sessionid)
  unit =  session.unit
  state = session.state

  // memorize sentence source (speech or text)  
  //asrFlag = data.speech

  // DEBUG
  //console.log('calling dialog:', unit, ' state: ', state)

  //
  // Call INPUT state function handler
  // pass also the user id as argument
  // https://stackoverflow.com/questions/8206453/call-function-by-string-name-on-node-js 
  Dialogs[unit].unit[state](sentence, sessionid)
}


/**
 * response
 *
 * send a message to user, through an external function
 * set timeout callback if arg is specificed
 *
 * @param {Object} TimeoutObject
 * @param {function} TimeoutObject.callback  function pointer
 * @param {ParamsObject} TimeoutObject.params parameters object 
 * @param {number} TimeoutObject.milliseconds number of timeout milliseconds 
 *
 * @param {String|DataObject} data - response data. Mandatory param
 * @param {TimeoutObject} timeout - optional param 
 * @param {String} sessionid - optional param
 *
 *
 * @see https://www.jstips.co/en/javascript/passing-arguments-to-callback-functions/
 * @return {function} externalResponse
 *
 */
function response(data, timeout, sessionid=getSessionId()) {

  const validatedData = validateResponseData(data)

  if (logFilename)
    log.out(sessionid, `[${stateid()}] ${validatedData.text}`)

  // start a timeout if argument 'timeout' is specificed
  if (timeout) {
    const { callback, params, milliseconds } = validateTimeout(timeout)

    startTimeout(callback, milliseconds, params, sessionid)
  }  

  // run the callback, after a data object validation
  return getExternalResponse()(sessionid, validatedData)
}


function prompt(data, timeout, sessionid=getSessionId()) {

  if ( !timeout )
    console.error(`WARNING: prompt(${JSON.stringify(data)}) without 'timeout' argument!`)

  return response(data, timeout, sessionid)
}  


// https://stackoverflow.com/a/3923258/1786393
module.exports = { 

  // out-dialog API
  stateid,
  currentState: stateid, // alias

  up,
  start,

  request,
  run: request, // alias

  down,

  // in-dialog API
  exec,
  goout: exec, // alias
  setOutputState: exec, // alias
  gotoOutputState: exec, // alias
  nextOutputState: exec, // alias

  next,
  goto: next, // alias
  goinp: next, // alias
  setInputState: next, // alias
  gotoInputState: next, // alias
  nextInputState: next, // alias
  
  end,
  exit: end, // alias
  quit: end, // alias

  getvar,
  setvar,
  delvar,
  delvars,
  
  logData,

  response,
  say: response, // alias
  reply: response, // alias

  prompt,
  ask: prompt,
  
  isSessionActive,
  isDialogActive: isSessionActive // alias
  
}

